/***********************************************************************************
                         anselm.ru [2014-08-21] GNU GPL
***********************************************************************************/

#include "daemon.h"
#include "log.h"

#include <libgen.h>   //basename
#include <sys/stat.h> //O_RDWR | O_EXCL | O_CREAT
#include <unistd.h>   //read
#include <signal.h>
#include <syslog.h>
#include <strings.h>  //bzero

void term(int signum) { //void term(int signum, siginfo_t *, void *) {
  debug( warning("stop signal %d", signum); )
  Daemon::alert_wake();
}

pthread_mutex_t Daemon::d_alert_mutex = NULL;//PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  Daemon::d_alert_cond  = NULL;//PTHREAD_COND_INITIALIZER;
bool            Daemon::d_is_alert    = false;

Daemon::Daemon(const char* binfile, const char* conffile, const char* pidfile) throw(const char*) {
  if(!binfile) {
    fprintf(stderr, "Error: variable \"APPNAME\" not defined");
    throw "binfile";
  }
  APPNAME = strdup(binfile);

  const char *TMP = conffile ? conffile : getenv("conffile");
  if(!TMP) {
    fprintf(stderr, "Error: variable environment \"CONFFILE\" not defined");
    throw "conffile";
  }
  CONFFILE = strdup(TMP);//strdup - обязательно!!
  
  const char *CONF = basename(CONFFILE);
  char *tmp = new char[strlen(APPNAME)+strlen(CONF)+1];
  sprintf(tmp, "%s %s", APPNAME, CONF);
  SYSLOGNAME = strdup(tmp);
  delete [] tmp;
  
  TMP = pidfile ? pidfile : getenv("pidfile");
  if(!tmp) {
    fprintf(stderr, "Error: variable environment \"PIDFILE\" not defined");
    throw "pidfile";
  }
  PIDFILE = strdup(TMP);//strdup - обязательно!!
}

Daemon::~Daemon() {
  delete [] APPNAME;
  delete [] CONFFILE;
  delete [] PIDFILE;
}

void
Daemon::childcode() {
  debug( warning("childcode - enter"); )
  //d_alert_mutex = PTHREAD_MUTEX_INITIALIZER;
  //d_alert_cond  = PTHREAD_COND_INITIALIZER; // вместо этого тонко настроим
  
  int err;
  pthread_mutexattr_t mutex_attr; 
  err = pthread_mutexattr_init(&mutex_attr);                                if(err) warning("pthread_mutexattr_init=%d", err);
  err = pthread_mutexattr_settype(&mutex_attr, PTHREAD_MUTEX_ERRORCHECK);   if(err) warning("pthread_mutexattr_settype=%d", err); //PTHREAD_MUTEX_NORMAL
  err = pthread_mutexattr_setpshared(&mutex_attr, PTHREAD_PROCESS_PRIVATE); if(err) warning("pthread_mutexattr_setpshared=%d", err);
  err = pthread_mutex_init(&d_alert_mutex, &mutex_attr);                    if(err) warning("pthread_mutex_init=%d", err);
  err = pthread_mutexattr_destroy(&mutex_attr);                             if(err) warning("pthread_mutexattr_destroy=%d", err);
  
  pthread_condattr_t cond_attr;
  err = pthread_condattr_init(&cond_attr);                                  if(err) warning("pthread_condattr_init=%d", err);
  err = pthread_condattr_setpshared(&cond_attr, PTHREAD_PROCESS_PRIVATE);   if(err) warning("pthread_condattr_setpshared=%d", err);
  err = pthread_condattr_setclock(&cond_attr, CLOCK_REALTIME);              if(err) warning("pthread_condattr_setclock=%d", err); //CLOCK_MONOTONIC "спасает от NTP"
  err = pthread_cond_init(&d_alert_cond, &cond_attr);                       if(err) warning("pthread_cond_init=%d", err);
  err = pthread_condattr_destroy(&cond_attr);                               if(err) warning("pthread_condattr_destroy=%d", err);

  childloop();
  
  pthread_cond_destroy(&d_alert_cond);
  pthread_mutex_destroy(&d_alert_mutex);  
  
  debug( warning("childcode - exit"); )
}


void
Daemon::childloop() {

// Example code (must be overloaded):
  debug( warning("childloop - enter"); )

  timeval tv = {1,0};
  
  while(1) {
    time_t t = time(NULL);
    warning("%d", t); //printf("%d\n", t);
    if(alert_sleep(tv)) break;// чутко спим до следующего опроса
  }
  
  debug( warning("childloop - exit"); )
}

bool
Daemon::start() {
  int f = open(PIDFILE, O_RDWR | O_EXCL | O_CREAT, S_IRUSR | S_IWUSR);
  if(f<0) {
    f = open(PIDFILE, O_RDONLY);
    char buf[21];
    read(f, buf, 20);
    int pid = atoi(buf);
    close(f);
    int i = kill(pid, 0);
    if(i<0) {
      remove(PIDFILE);
      f = open(PIDFILE, O_RDWR | O_TRUNC | O_CREAT, S_IRUSR | S_IWUSR);
    }
    else {
      fprintf(stderr, "%s (pid=%d) already running\n", APPNAME, pid);
      return false;
    }
  }
  
  int pid;
  pid = fork();
  switch(pid) {
  case 0:
    close(f);
    setsid();
    chdir("/");
    close(0);
    close(1);
    close(2);
    
    struct sigaction act;
    bzero(&act, sizeof(act));
    act.sa_handler = term; //act.sa_sigaction = term;
    sigaction(SIGTERM, &act, NULL); //signal(SIGTERM, term);
    //signal(SIGHUP,  reload);
    signal(SIGPIPE, SIG_IGN);  
    childcode();
    remove(PIDFILE);
    
    return true;
  case -1:
    fprintf(stderr, "Error: unable to fork\n");
    break;
  default:
    char buf[21] = {0};
    sprintf(buf, "%d\n", pid);
    write(f, buf, 20);
    break;
  }
    
  close(f);
  
  return true;
}

//ф-я чуткого сна
bool
Daemon::alert_sleep(const timeval& tv) {
  int err = pthread_mutex_lock(&d_alert_mutex); // перед вызовом pthread_cond_wait мютекс должен быть заблокирован
  if(err!=0) return false;
  
  timeval now;
  gettimeofday(&now, DST_NONE);
  struct timespec ts;
  ts.tv_sec  = now.tv_sec + tv.tv_sec;
  ts.tv_nsec = (now.tv_usec + tv.tv_usec) * 1000; // usec to nsec
  int ret = pthread_cond_timedwait(&d_alert_cond, &d_alert_mutex, &ts);
  pthread_mutex_unlock(&d_alert_mutex);
  return ret==0; // дождались ли ответа вовремя//60 != ETIMEDOUT
}

// ф-я пробуждения
bool
Daemon::alert_wake() {
  int ret = pthread_cond_signal(&d_alert_cond);
  if(ret == -1) warning("Error: pthread_cond_signal=-1");
  d_is_alert = true;
  return ret==0;
}

// проверка условия
bool
Daemon::alert_check() {
  return d_is_alert;
}
